#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 22 10:23:02 2019

@author: bleyerj
"""
import dolfinx
from dolfinx import fem
from dolfinx.fem.petsc import LinearProblem, NonlinearProblem
from dolfinx.io import XDMFFile
from ufl import (
    derivative,
    dot,
    Measure,
    grad,
    sym,
    inner,
    tr,
    Identity,
    TrialFunction,
    TestFunction,
    dx,
    dev,
)
from dolfinx.nls.petsc import NewtonSolver
import warnings
import numpy as np
from mpi4py import MPI
from petsc4py import PETSc
from plotting import plot_damage, create_grid
from tao_problem import TAOProblem
import logging
from mealor import DirichletBoundaryCondition
from mealor.utils import mark_facets

logging.disable(logging.INFO)

linear_solver = "lu"  # or "mumps"


def solve_problem(
    domain,
    subdomains,
    facets,
    prob_params,
    mech_params,
):
    unloading, export_results, Nitermax, tol = prob_params
    E, nu, Gc, l0, model, Umax, Nincr = mech_params

    dx = Measure("dx", domain=domain, subdomain_data=subdomains)

    # stiffness penalty for rigid pins
    kpen = fem.Constant(domain, 100.0)

    lmbda = fem.Constant(domain, E * nu / (1 + nu) / (1 - 2 * nu))
    mu = fem.Constant(domain, E / 2 / (1 + nu))
    lmbda_ps = (
        2 * mu * lmbda / (lmbda + 2 * mu)
    )  # effective lambda modulus for plane stress
    kres = fem.Constant(domain, 1e-6)  # residual stiffness
    l0 = fem.Constant(domain, l0)
    Gc = fem.Constant(domain, Gc)

    kappa = lmbda_ps + 2 / 3 * mu

    def ppos(v):
        return (v + abs(v)) / 2

    def pneg(v):
        return (-v + abs(v)) / 2

    def eps(v):
        return sym(grad(v))

    def sigma0(v):
        return lmbda_ps * tr(eps(v)) * Identity(2) + 2 * mu * eps(v)

    def g(d):
        return (1 - d) ** 2 + kres

    def sigma(v, d):
        return g(d) * sigma0(v)

    def psi_pos(v):
        ed = dev(eps(v))
        return 0.5 * ppos(tr(eps(v))) ** 2 + mu * inner(ed, ed)

    def psi_neg(v):
        return 0.5 * pneg(tr(eps(v))) ** 2

    if unloading:
        load_steps = np.concatenate(
            (
                np.linspace(0, 3 * Umax / 4, Nincr // 2 + 1),
                np.linspace(3 * Umax / 4, 0, 20)[1:],
                np.linspace(0, Umax, Nincr // 2 + 1)[1:],
            )
        )
    else:
        load_steps = np.linspace(0, Umax, Nincr + 1)

    # function space for the displacement
    V_u = fem.VectorFunctionSpace(domain, ("CG", 1))
    # function space for the damage
    V_d = fem.FunctionSpace(domain, ("CG", 1))
    # function space for the stress
    V_sig = fem.TensorFunctionSpace(domain, ("DG", 0))

    tol = 0.05

    # def top_pin(x):
    #     return np.logical_and(np.isclose(x[0], 12.0, tol), np.isclose(x[1], 12, tol))

    # def bot_pin(x):
    #     return np.logical_and(np.isclose(x[0], 12.0, tol), np.isclose(x[1], -12, tol))

    def right(x):
        return np.isclose(x[0], 1)

    def left(x):
        return np.isclose(x[0], 0)

    facets = mark_facets(domain, {1: right, 3: left})
    ds = Measure("ds", domain=domain, subdomain_data=facets)
    # top_pin_dofs = locate_dofs_topological(
    #     V_u.sub(1), 2, subdomains.indices[subdomains.values == 2]
    # )
    # bot_pin_dofs = locate_dofs_topological(
    #     V_u.sub(1), 2, subdomains.indices[subdomains.values == 3]
    # )
    # V_ux, mapping = V_u.sub(0).collapse()
    # V_uy, mapping = V_u.sub(1).collapse()
    # top_pin_dofs_x = locate_dofs_geometrical((V_u.sub(0), V_ux), top_pin)
    # bot_pin_dofs_x = locate_dofs_geometrical((V_u.sub(0), V_ux), bot_pin)
    # right_dofs_x = locate_dofs_geometrical((V_u.sub(0), V_ux), right)
    # top_pin_dofs_y = locate_dofs_geometrical((V_u.sub(1), V_uy), top_pin)
    # bot_pin_dofs_y = locate_dofs_geometrical((V_u.sub(1), V_uy), bot_pin)

    # top_pin_dofs_x = [np.array([top_pin_dofs_x[i][0]]) for i in range(2)]
    # bot_pin_dofs_x = [np.array([bot_pin_dofs_x[i][0]]) for i in range(2)]
    # # right_dofs_x = locate_dofs_topological(
    # #     V_u.sub(0), facets.dim, facets.indices[facets.values == 1]
    # # )
    # U0 =fem.Function(V_ux)
    # Uimpt =fem.Function(V_uy)
    # Uimpb =fem.Function(V_uy)
    # bcs = [
    #     dirichletbc(Uimpt, top_pin_dofs_y, V_u.sub(1)),
    #     dirichletbc(Uimpb, bot_pin_dofs_y, V_u.sub(1)),
    #     # dirichletbc(U0, top_pin_dofs_x, V_u.sub(0)),
    #     # dirichletbc(U0, bot_pin_dofs_x, V_u.sub(0)),
    #     dirichletbc(U0, right_dofs_x, V_u.sub(0)),
    # ]
    Uimp = fem.Constant(domain, 0.0)
    dirichlet = DirichletBoundaryCondition(V_u)
    dirichlet.add_bc_topological(facets, 1, ux=Uimp)
    dirichlet.add_bc_topological(facets, 3, uy=0, ux=0)

    u = fem.Function(V_u, name="Total_displacement")
    v = TrialFunction(V_u)
    u_ = TestFunction(V_u)
    sig = fem.Function(V_sig, name="Current_stress")

    d = fem.Function(V_d, name="Damage")
    dold = fem.Function(V_d, name="Previous_damage")
    dub = fem.Function(V_d, name="Upper_bound_d=1")
    with dub.vector.localForm() as bc_local:
        bc_local.set(1.0)
    dlb = fem.Function(V_d, name="Lower_bound_d_n")
    d_ = TestFunction(V_d)
    dd = TrialFunction(V_d)
    # boundary conditions for damage problem
    # top_dofs = fem.locate_dofs_topological(
    #     V_d, 2, subdomains.indices[subdomains.values == 2]
    # )
    # bot_dofs = locate_dofs_topological(
    #     V_d, 2, subdomains.indices[subdomains.values == 3]
    # )
    # bc_d = [dirichletbc(0.0, top_dofs, V_d), dirichletbc(0.0, bot_dofs, V_d)]

    # set_bc(dub.vector, bc_d)

    # bilinear and l form for elasticity problem
    a = inner(eps(v), sigma(u_, dold)) * dx
    l = fem.Constant(domain, 0.0) * u_[0] * dx
    # linear elastic problem at fixed damage
    problem_u = LinearProblem(
        a,
        l,
        bcs=dirichlet.bcs,
        u=u,
        petsc_options={"ksp_type": "preonly", "pc_type": linear_solver},
    )

    # total energy for the damage problem
    # elastic_energy = (g(d) * psi_pos(u) + psi_neg(u)) * dx
    elastic_energy = g(d) * 0.5 * inner(eps(u), sigma0(u)) * dx
    if model == "AT1":
        cw = fem.Constant(domain, 8 / 3.0)
        w = lambda d: d
    elif model == "AT2":
        cw = fem.Constant(domain, 2.0)
        w = lambda d: d**2
    fracture_energy = Gc / cw * (w(d) / l0 + l0 * dot(grad(d), grad(d))) * dx
    total_energy = elastic_energy + fracture_energy
    F_u = derivative(elastic_energy, u, v)
    J_u = derivative(F_u, u, u_)

    problem_u = NonlinearProblem(
        F_u,
        u,
        bcs=dirichlet.bcs,
        J=J_u,
    )
    u_solver = NewtonSolver(domain.comm, problem_u)

    u_solver.convergence_criterion = "incremental"
    u_solver.rtol = 1e-6
    u_solver.report = True

    # if bcs_damage:
    #     left_dofs = locate_dofs_topological(
    #         V_d, facets.dim, facets.indices[facets.values == 1]
    #     )
    #     right_dofs = locate_dofs_topological(
    #         V_d, facets.dim, facets.indices[facets.values == 2]
    #     )
    #     bc_d = [dirichletbc(0.0, left_dofs, V_d), dirichletbc(0.0, right_dofs, V_d)]
    # else:
    bc_d = []

    # first derivative of energy with respect to d
    F_dam = derivative(total_energy, d, d_)
    # second derivative of energy with respect to d
    J_dam = derivative(F_dam, d, dd)
    # Definition of the optimisation problem with respect to d
    damage_problem = TAOProblem(total_energy, F_dam, J_dam, d, bc_d)

    b = dolfinx.la.create_petsc_vector(V_d.dofmap.index_map, V_d.dofmap.index_map_bs)
    J = dolfinx.fem.petsc.create_matrix(damage_problem.a)

    # Create PETSc TAO
    solver_d_tao = PETSc.TAO().create()
    solver_d_tao.setType("tron")
    solver_d_tao.setObjective(damage_problem.f)
    solver_d_tao.setGradient(damage_problem.F, b)
    solver_d_tao.setHessian(damage_problem.J, J)
    solver_d_tao.setTolerances(grtol=1e-6, gttol=1e-6)
    solver_d_tao.view()

    solver_d_tao.getKSP().setType("preonly")
    solver_d_tao.getKSP().setTolerances(rtol=1.0e-6)
    solver_d_tao.getKSP().getPC().setType("lu")

    Nincr = len(load_steps) - 1
    ffile = XDMFFile(MPI.COMM_WORLD, "results.xdmf", "w")
    ffile.write_mesh(domain)

    vol = fem.assemble_scalar(fem.form(fem.Constant(domain, 1.0) * dx))
    results = np.zeros((Nincr + 1, 4))
    t = 0
    for i, t in enumerate(load_steps[1:]):
        Uimp.value = t
        dirichlet.update()

        print("Increment {:3d}".format(i + 1))

        # We set the bound (Note: they are passed as reference and not as values)
        solver_d_tao.setVariableBounds(dlb.vector, dub.vector)

        niter = 0
        for niter in range(Nitermax):
            # Solve displacement
            # problem_u.solve()
            u_solver.solve(u)
            # Compute new damage
            solver_d_tao.solve(d.vector)

            # check error and update
            L2_error = fem.form(inner(d - dold, d - dold) * dx)
            error_L2 = np.sqrt(fem.assemble_scalar(L2_error)) / vol

            # Update damage
            d.vector.copy(dold.vector)

            print("    Iteration {:3d}: ||Res||_2={:5e}".format(niter, error_L2))
            if error_L2 < tol:
                break
        else:
            warnings.warn("Too many iterations in fixed point algorithm")

        # Update lower bound to account for irreversibility
        d.vector.copy(dlb.vector)

        # compute apparent stress and max damage
        # if problem == "shear":
        #     L = assemble_scalar(form(fem.Constant(mesh, 1.0) * ds(4)))
        #     Sig_app = assemble_scalar(form(sigma(u, d)[0, 1] * ds(4))) / L
        # else:
        #     W = assemble_scalar(form(fem.Constant(mesh, 1.0) * ds(2)))
        #     Sig_app = assemble_scalar(form(sigma(u, d)[0, 0] * ds(2))) / W

        # en_el = assemble_scalar(form(elastic_energy)) / Gc.value
        # en_f = assemble_scalar(form(fracture_energy)) / Gc.value

        # results[i + 1, :] = (t, Sig_app, en_el, en_f)

        # compute stress
        stress_expr = fem.Expression(sigma(u, d), V_sig.element.interpolation_points())
        sig.interpolate(stress_expr)

        # Create plotter
        # if PYVISTA_ON and (
        #     (plot_every > 0 and i % plot_every == 0)
        #     or (plot_every == 0 and i == Nincr - 1)
        # ):
        #     plot_damage(grid, d)

        # Export to Paraview
        if export_results:
            ffile.write_function(d, t)
            ffile.write_function(u, t)
            # ffile.write_function(sig, t)

        ffile.close()
    return results, u, d, sig
